use std::{collections::HashMap, borrow::Cow};
use clap::Parser;
use serde_derive::Deserialize;
use colored::Colorize;

macro_rules! args_or_config {
    ($(#[$($attr: tt)*] $val: ident: $t: ty),* |optional: $(#[$($attr_opt: tt)*] $val_opt: ident: $t_opt: ty),*) => {
        #[derive(Parser, Debug)]
        #[clap(version)]
        struct ClapArgs {
            $(
                #[$($attr)*]
                $val: Option<$t>,
            )*
            $(
                #[$($attr_opt)*]
                $val_opt: Option<$t_opt>,
            )*
            #[clap(help("Host filter used for finding newly added devices in the device management. The host filter can be retrieved by performing a search in the device admin section and copying the query from the url. On the command line each filters key and value are seperated by a colon."))]
            host_filter: Vec<String>
        }
        #[derive(Debug, Deserialize)]
        struct TomlArgs {
            $( $val: Option<$t>, )*
            $( $val_opt: Option<$t_opt>, )*
            host_filter: HashMap<String, String>
        }

        impl ClapArgs {
            fn combine(self, toml: &str) -> Config {
                let toml: TomlArgs = toml::from_str(toml).expect("Invalid config");

                Config {
                    $( 
                        $val: self.$val.unwrap_or_else(||
                            toml.$val.unwrap_or_else(|| {
                                eprintln!("{}{}{}{}{}{}",
                                    "Missing config value for".bright_red(),
                                    " '",
                                    stringify!($val).cyan(),
                                    "'",
                                    "\nHINT: ".green(),
                                    "Specify it either in the config.toml or as an argument.".yellow()
                                );
                                std::process::exit(187);
                            })
                        ),
                    )*
                    $( $val_opt: self.$val_opt.or(toml.$val_opt), )*
                    host_filter: self.host_filter.into_iter()
                        .map(|s| {
                            let (l, r) = s.split_once(':').expect("Params need ':' between key and value");
                            (l.to_owned(), r.to_owned())
                        })
                        .chain(toml.host_filter)
                        .collect()
                }
            }
        }

        struct Config {
            $( $val: $t, )*
            $( $val_opt: Option<$t_opt>, )*
            host_filter: HashMap<String, String>
        }
        impl Config {
            fn parse(toml: &str) -> Self {
                ClapArgs::parse().combine(toml)
            }
        }

    };
}

args_or_config! {
    #[clap(short('H'), long, help("Hostname where the IServ shall be reached at"))]
    hostname: String,
    #[clap(short, long, help("Cookie used for authentication. 'IServSAT' and 'IServSession' are required. HINT: You can retrieve the cookie data by using the developer tools of your browser."))]
    cookie: String,
    #[clap(long, help("Preferred ip address for the newly added device. If the preferred ip address is already used, the next available ip address is used instead."))]
    host_ip: String,
    #[clap(long, help("Numerical room id the newly added device should be assigned to. HINT: The id can be found at the end of the url when inspecting a room in the room admin section."))]
    host_room: u16,
    #[clap(long, help("Amount of time in seconds to wait for the next retry."))]
    retry_delay: u64

    |optional:

    #[clap(long, help("Numerical compilation id the newly added device should be assigned to. HINT: The id can be found at the end of the url when inspecting a compilation in the mdm admin section. If this is not specified, the new device will not be added to a collection."))]
    mdm_compilation: u16
}

macro_rules! print_banner {
    ($f: expr, $($arg: expr),*) => {
        println!(concat!(
            "----------------------------------------\n",
            $f,
            "\n----------------------------------------"
        ), $($arg),*)
    };
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let args = Config::parse(&std::fs::read_to_string("config.toml").unwrap_or(String::new()));
    loop { main_loop(&args)?; }
}

fn main_loop(args: &Config) -> Result<(), Box<dyn std::error::Error>> {
    let header = ("Cookie", &args.cookie);
    let client = reqwest::blocking::Client::new();
    let resp = client
        .get(format!("https://{}/iserv/admin/hosts", args.hostname))
        .header(header.0, header.1)
        .query(&args.host_filter)
        .send()?;
    let status = &resp.status();
    let plain = &resp.text()?;

    if plain.contains("class=\"login-form\"") {
        eprintln!("ERR: Failed to authorize. Server forwarded to login site. HINT: Check the provided cookie.");
        std::process::exit(187);
    }
    
    if status.is_client_error() || status.is_server_error() {
        eprintln!("ERR: Received error status (status code: {:?})", status);
        std::process::exit(187);
    }
    
    let content = squeeze(
        &plain,
        "<script id=\"crud-data\" type=\"application/json\">",
        "</script>",
    );
    let json: serde_json::Value = serde_json::from_str(content).expect("ERR: Failed to parse JSON data. This may be because the server did not return valid JSON data to parse.");
    let json_part = &json["data"][0]["name"]["rendered"]
        .to_string()
        .replace("\\", "");
    // skip rest of loop when no entry appeared

    if json_part == "null" {
        println!(
"########################################
No new host: skipping (Retry in {}s)
########################################
", &args.retry_delay);
        std::thread::sleep(std::time::Duration::from_secs(5));
        return Ok(());
    }

    let host_id = squeeze(json_part, "data-host-id=\"", "\"");
    let device_name = squeeze(json_part, "</span>", "</a>");
    print_banner!("1/8 Found host\nHost id: {}\nDevice name: {}\nReceived response status: {:?}", host_id, device_name, status);
    
    let dest = format!("https://{}/iserv/admin/host/edit/{}", args.hostname,  &host_id);
    let resp = client.get(&dest).header(header.0, header.1).send()?;
    let status = &resp.status();
    if !status.is_success() {
        eprintln!("{}, {}", "WARN: Found invalid Status Code: ".bright_yellow(), status);
        return Ok(())
    }
    let plain = &resp.text()?;
    let mut form_ip: &str = &args.host_ip;
    let mut form_ip_index = 2;
    let form_name = squeeze_html_form(plain, "name=\"host[name]\"", ">", "value=\"", "\"");
    let plain_inner_tags = squeeze(plain, "name=\"host[tags][]\"", "</select>");
    let mut form_tags = Vec::new();
    squeeze_loop(&mut form_tags, plain_inner_tags, "<option value=\"", "\"", " selected=\"selected\"", 3, "host[tags][]");
    
    let form_mac = squeeze_html_form(plain, "name=\"host[mac]\"", ">", "value=\"", "\"");
    let form_description = squeeze_html_form(plain, "name=\"host[description]\"", "/textarea>", ">", "<");
    let form_token = squeeze_html_form(plain, "name=\"host[_token]\"", ">", "value=\"", "\"");
    let mut form_data = Vec::from([
        ("host[name]",form_name),
    ].map(|(a, b)| (Cow::Borrowed(a), Cow::Borrowed(b))));
        
    // empty tag fields are not permitted
    if !(form_tags.is_empty()) {
        form_ip_index += form_tags.len();
        form_data.extend(form_tags);
    }
        
    let host_room: &str = &args.host_room.to_string();
    form_data.extend([    
    ("host[room]", host_room),
    ("host[ip]", form_ip),
    ("host[mac]", form_mac),
    ("host[internet]", "1"),
    ("host[proxyEnforce]", "0"),
    ("host[owner]", ""),
    ("host[controllable]", "0"),
    ("host[inventoryNumber]", ""),
    ("host[shutdown]", ""),
    ("host[description]", form_description),
    ("host[image][id]", ""),
    ("host[image][x]", ""),
    ("host[image][y]", ""),
    ("host[actions][submit]", ""),
    ("host[_token]", form_token),
    ].map(|(a, b)| (Cow::Borrowed(a), Cow::Borrowed(b))));

    print_banner!("2/8 Received current host properties\nReceived response status: {:?}", status);
        
    println!("{:#?}", form_data);
    let resp = client
        .post(&dest)
        .header(header.0, header.1)
        .form(&form_data)
        .send()?;
    let status = &resp.status();
    if !status.is_success() {
        eprintln!("{}, {}", "WARN: Found invalid Status Code: ".bright_yellow(), status);
        return Ok(())
    }
    let plain = &resp.text()?;


    if !plain.contains("<form name=\"host\"") {
        print_banner!("3/8 The configured ip was accepted {}\nReceived response status: {:?}", form_ip, status);
        print_banner!("{}", "4/8 Ip address resubmission was skipped");
    } else {
        // retrieve next available ip address from sent form
        form_ip = squeeze_html_form(plain, "name=\"host[ip]\"", ">", "value=\"", "\"");
        form_data[form_ip_index] = ("host[ip]".into(), form_ip.into());
        print_banner!("3/8 Received recommended ip: {}\nReceived response status: {:?}", form_ip, status);
        
        let resp = client
            .post(&dest)
            .header(header.0, header.1)
            .form(&form_data)
            .send()?;
        let status = &resp.status();
        print_banner!("4/8 Sucessfully updated host\nReceived response status: {:?}", status);
    }

    let target_compilation = if let Some(t) = args.mdm_compilation { t } else { return Ok(()) };
    
    let resp = client
        .get(format!("https://{}/iserv/admin/mdm/ios/compilation/edit/{}", args.hostname, target_compilation))
        .header(header.0, header.1)
        .send()?;
    
    let status = &resp.status();
    if !status.is_success() {
        eprintln!("{}, {}", "WARN: Found invalid Status Code: ".bright_yellow(), status);
        return Ok(())
    }
    let plain = &resp.text()?;
    
    print_banner!("5/8 Received MDM compilation\nReceived response status: {:?}", status);

    let mut form_data = Vec::from([
        (
            "ioscompilation[name]",
            squeeze_html_form(plain, "name=\"ioscompilation[name]\"", ">", "value=\"", "\"")
        ),
        (
            "ioscompilation[description]",
            squeeze_html_form(plain, "name=\"ioscompilation[description]\"", ">", "value=\"", "\"")
        ),
    ].map(|(a, b)| (Cow::Borrowed(a), Cow::Borrowed(b))));
    
    // push apps
    let plain_inner_applications = squeeze(plain, "name=\"ioscompilation[applications][]\"", "</select>");
    squeeze_loop(&mut form_data, plain_inner_applications, "<option value=\"", "\"", " selected=\"selected\"", 3, "ioscompilation[applications][]");

    // push profiles
    let plain_inner_profiles = squeeze(plain, "name=\"ioscompilation[profiles][]\"", "</select>");
    squeeze_loop(&mut form_data, plain_inner_profiles, "<option value=\"", "\"", " selected=\"selected\"", 3, "ioscompilation[profiles][]");

    // push existing devices
    let plain_inner_devices = squeeze(plain, "name=\"ioscompilation[devices][]\"", "</select>");
    squeeze_loop(&mut form_data, plain_inner_devices, "<option value=\"", "\"", " selected=\"", 3, "ioscompilation[devices][]");

    // push new device
    let mdm_new_idevice_id = squeeze_right(plain_inner_devices, "<option value=\"", &format!("\">{}", form_name));
    if plain_inner_devices.contains(form_name) {
        form_data.push(("ioscompilation[devices][]".into(), mdm_new_idevice_id.into()));
    } else {
        eprintln!("{}", "WARN: Device not found in compilation form.\nSkipping compilation edit...".bright_yellow());
        return Ok(());
    }

    // push submit action
    form_data.push(("ioscompilation[actions][submit]".into(), "".into()));

    //push token
    let form_token = squeeze_html_form(plain, "name=\"ioscompilation[_token]\"", ">", "value=\"", "\"");
    form_data.push(("ioscompilation[_token]".into(), form_token.into()));
    let resp = client
    .post(format!("https://{}/iserv/admin/mdm/ios/compilation/edit/{}", args.hostname, args.mdm_compilation.unwrap()))
        .header(header.0, header.1)
        .form(&form_data)
        .send()?;
    
    let status = &resp.status();
    if !status.is_success() {
        eprintln!("{}, {}", "WARN: Found invalid Status Code: ".bright_yellow(), status);
        return Ok(())
    }
    print_banner!("6/8 Added device to MDM Compilation\nReceived response status: {:?}", status);

    let resp = client
        .get(format!("https://{}/iserv/admin/mdm/ios/device", args.hostname))
        .header(header.0, header.1)
        .send()?;

    let status = &resp.status();
    if !status.is_success() {
        eprintln!("{}, {}", "WARN: Found invalid Status Code: ".bright_yellow(), status);
        return Ok(())
    }
    let plain = &resp.text()?;
    let csrf_token = squeeze_html_form(plain, "name=\"iserv_crud_multi_select[_token]\"", ">", "value=\"", "\"");

    print_banner!("7/8 Retrieved CSRF token\nReceived response status: {:?}", status);
        
    let form_data = [
        ("iserv_crud_multi_select[confirm]", "apply-changes"),
        ("iserv_crud_multi_select[multi][]", mdm_new_idevice_id),
        ("iserv_crud_multi_select[topReplacementActions][apply-changes]", ""),
        ("iserv_crud_multi_select[_token]", csrf_token),
    ];
    let resp = client
        .post(format!("https://{}/iserv/admin/mdm/ios/device/batch", args.hostname))
        .header(header.0, header.1)
        .form(&form_data)
        .send()?;
    let status = &resp.status();
    if !status.is_success() {
        eprintln!("{}, {}", "WARN: Found invalid Status Code: ".bright_yellow(), status);
        return Ok(())
    }
    print_banner!("8/8 Confirmed device actions\nReceived response status: {:?}", status);
    Ok(())
}

// helper functions (needed for naive html parsing)

fn squeeze<'inp>(input: &'inp str, before: &str, after: &str) -> &'inp str {
    //find before occurence (excluded from output)
    if let Some(before_bytes) = input.find(before) {
        let adjusted_before_bytes = before_bytes + before.len();
        let leftover = &input[adjusted_before_bytes..];
        //find after occurence (excluded from output)
        if let Some(after_bytes) = leftover.find(after) {
            let adjusted_after_bytes = after_bytes + adjusted_before_bytes;
            // ensure bytes are not equal
            if adjusted_before_bytes < adjusted_after_bytes {
                return &input[adjusted_before_bytes..adjusted_after_bytes];
            }
        }
    }
    ""
}

fn squeeze_right<'inp>(input: &'inp str, before: &str, after: &str) -> &'inp str {
    //find before occurence (excluded from output)
    if let Some(after_bytes) = input.find (after) {
        let leftover = &input[..after_bytes];
        //find after occurence (excluded from output)
        if let Some(before_bytes) = leftover.rfind(before) {
            let adjusted_before_bytes = before_bytes + before.len();
            // ensure bytes are not equal
            if adjusted_before_bytes < after_bytes {
                return &input[adjusted_before_bytes..after_bytes];
            }
        }
    }
    ""
}

fn squeeze_html_form<'inp>(
    input: &'inp str,
    outer_before: &str,
    outer_after: &str,
    inner_before: &str,
    inner_after: &str
) -> &'inp str {
    let outer_str = squeeze(input, outer_before, outer_after);
    squeeze(outer_str, inner_before, inner_after)
}

fn squeeze_loop<'inp>(
    collector: &mut Vec<(Cow<'inp, str>, Cow<'inp, str>)>,
    input: &'inp str,
    before: &str,
    after: &str,
    must_include_after: &str,
    include_after_distance: usize,
    index_name: &'inp str
)  {
    let mut leftover = input;
    //find before occurence (excluded from output)
    while let Some(before_bytes) = leftover.find(before) {
        let input = leftover;
        let adjusted_before_bytes = before_bytes + before.len();
        leftover = &leftover[adjusted_before_bytes..];
        //find after occurence (excluded from output)
        if let Some(after_bytes) = leftover.find(after) {
            let absolute_after_bytes = after_bytes + adjusted_before_bytes;
            leftover = &leftover[after_bytes + after.len()..];
            // ensure bytes are not equal and that the item is "selected"/enabled
            if let Some(must_include_distance) = leftover.find(must_include_after) {
                if adjusted_before_bytes < absolute_after_bytes && must_include_distance < include_after_distance {
                    collector.push((index_name.into(), input[adjusted_before_bytes..absolute_after_bytes].into()));
                }
            }
        }
    }
}